"""
Absfuyu: Passwordlib
--------------------
Password library

Version: 1.0.0
Date updated: 11/02/2025 (dd/mm/yyyy)
"""

# Module level
###########################################################################
__all__ = ["PasswordGenerator"]


# Library
###########################################################################
import hashlib
import lzma
import os
import random
import re
from typing import NamedTuple, Optional

from absfuyu.general.data_extension import DictExt, Text
from absfuyu.general.generator import Charset, Generator
from absfuyu.logger import logger
from absfuyu.pkg_data import DataList
from absfuyu.util import set_min
from absfuyu.util.pkl import Pickler


# Function
###########################################################################
def _password_check(password: str) -> bool:
    """
    Verify the strength of ``password``.
    A password is considered strong if:

    - 8 characters length or more
    - 1 digit or more
    - 1 symbol or more
    - 1 uppercase letter or more
    - 1 lowercase letter or more

    :param password: Password want to be checked
    :type password: str
    :rtype: bool
    """

    # calculating the length
    length_error = len(password) < 8

    # searching for digits
    digit_error = re.search(r"\d", password) is None

    # searching for uppercase
    uppercase_error = re.search(r"[A-Z]", password) is None

    # searching for lowercase
    lowercase_error = re.search(r"[a-z]", password) is None

    # searching for symbols
    symbols = re.compile(r"[ !#$%&'()*+,-./[\\\]^_`{|}~" + r'"]')
    symbol_error = symbols.search(password) is None

    detail = {
        "length_error": length_error,
        "digit_error": digit_error,
        "uppercase_error": uppercase_error,
        "lowercase_error": lowercase_error,
        "symbol_error": symbol_error,
    }
    logger.debug(f"Password error summary: {detail}")

    return not any(
        [
            length_error,
            digit_error,
            uppercase_error,
            lowercase_error,
            symbol_error,
        ]
    )


# Class
###########################################################################
class PasswordHash(NamedTuple):
    salt: bytes
    key: bytes


class PasswordGenerator:
    """Password Generator"""

    def __str__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __repr__(self) -> str:
        return self.__str__()

    @staticmethod
    def password_hash(password: str) -> PasswordHash:
        """
        Generate hash for password
        """
        salt = os.urandom(32)
        key = hashlib.pbkdf2_hmac(
            hash_name="sha256",
            password=password.encode("utf-8"),
            salt=salt,
            iterations=100000,
        )
        # out = {
        #     "salt": salt,
        #     "key": key,
        # }
        out = PasswordHash(salt, key)
        return out

    @staticmethod
    def password_check(password: str) -> dict:
        data = Text(password).analyze()
        data = DictExt(data).apply(lambda x: True if x > 0 else False)  # type: ignore
        data.__setitem__("length", len(password))
        return dict(data)

    # Password generator
    @staticmethod
    def generate_password(
        length: int = 8,
        include_uppercase: bool = True,
        include_number: bool = True,
        include_special: bool = True,
    ) -> str:
        r"""
        Generate a random password

        Parameters
        ----------
        length : int
            | Length of the password.
            | Minimum value: ``8``
            | (Default: ``8``)

        include_uppercase : bool
            Include uppercase character in the password (Default: ``True``)

        include_number : bool
            Include digit character in the password (Default: ``True``)

        include_special : bool
            Include special character in the password (Default: ``True``)

        Returns
        -------
        str
            Generated password


        Example:
        --------
        >>> Password.generate_password()
        [T&b@mq2
        """
        charset = Charset.LOWERCASE
        check = 0

        if include_uppercase:
            charset += Charset.UPPERCASE
            check += 1

        if include_number:
            charset += Charset.DIGIT
            check += 1

        if include_special:
            charset += r"[ !#$%&'()*+,-./[\\\]^_`{|}~" + r'"]'
            check += 1

        while True:
            pwd = Generator.generate_string(
                charset=charset,
                size=set_min(length, min_value=8),  # type: ignore
                times=1,
                string_type_if_1=True,
            )

            analyze = Text(pwd).analyze()  # Count each type of char

            s = sum([1 for x in analyze.values() if x > 0])  # type: ignore
            if s > check:  # Break loop if each type of char has atleast 1
                break
        return pwd  # type: ignore

    @staticmethod
    def generate_passphrase(
        num_of_blocks: int = 5,
        block_divider: Optional[str] = None,
        first_letter_cap: bool = True,
        include_number: bool = True,
        *,
        custom_word_list: list[str] | None = None,
    ) -> str:
        """
        Generate a random passphrase

        Parameters
        ----------
        num_of_blocks : int
            Number of word used (Default: ``5``)

        block_divider : str
            Character symbol that between each word (Default: ``"-"``)

        first_letter_cap : bool
            Capitalize first character of each word (Default: ``True``)

        include_number : bool
            Add number to the end of each word (Default: ``True``)

        custom_word_list : list[str] | None
            Custom word list for passphrase generation, by default uses a list of 360K+ words

        Returns
        -------
        str
            Generated passphrase


        Example:
        --------
        >>> print(Password().generate_passphrase())
        Myomectomies7-Sully4-Torpedomen7-Netful2-Begaud8
        """
        words: list[str] = (
            (lzma.decompress(Pickler.load(DataList.PASSWORDLIB)).decode().split(","))  # type: ignore
            if not custom_word_list
            else custom_word_list
        )

        if not block_divider:
            block_divider = "-"

        dat = [random.choice(words) for _ in range(num_of_blocks)]

        if first_letter_cap:
            dat = list(map(lambda x: x.title(), dat))

        if include_number:
            idx = random.choice(range(5))
            dat[idx] += str(random.choice(range(10)))

        return block_divider.join(dat)


# Run
###########################################################################
if __name__ == "__main__":
    logger.setLevel(10)
